#!/bin/bash

log_base=$($CLI --print-logdir)
log_id=${B0}/${V0}-0
log_id=${log_id:1}     # Remove initial slash
log_id=${log_id//\//-} # Replace remaining slashes with dashes
FDL_META_FILE=${log_base}/${log_id}-meta-1.jnl
FDL_DATA_FILE=${log_base}/${log_id}-data-1.jnl

check_logfile() {
        [ $(gf_logdump $FDL_META_FILE $FDL_DATA_FILE | grep $1 | wc -l) -ge $2 ]
}
