#!/bin/bash

# Test if only a single xml document is generated by 'status all'
# when a volume is not started

. $(dirname $0)/../../include.rc

cleanup;

TEST glusterd

TEST $CLI volume create ${V0}1 $H0:$B0/${V0}1{1,2}
TEST $CLI volume create ${V0}2 $H0:$B0/${V0}2{1,2}

TEST $CLI volume start ${V0}1

function test_status_all ()
{
        $CLI volume status all --xml | xmllint -format -
}

TEST test_status_all

TEST $CLI volume stop ${V0}1

cleanup
